@@ -3,26 +3,25 @@
  * functions.c
  *	  Execution of SQL-language functions
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/executor/functions.c,v 1.91.4.3 2007/02/02 00:03:44 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/executor/functions.c,v 1.108.2.2 2007/04/02 18:49:36 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
-#include "access/heapam.h"
+#include "access/xact.h"
 #include "catalog/pg_proc.h"
 #include "catalog/pg_type.h"
 #include "commands/trigger.h"
-#include "executor/executor.h"
 #include "executor/functions.h"
+#include "funcapi.h"
 #include "parser/parse_coerce.h"
 #include "parser/parse_expr.h"
-#include "parser/parse_type.h"
 #include "tcop/tcopprot.h"
 #include "tcop/utility.h"
 #include "utils/builtins.h"
@@ -81,7 +80,7 @@ typedef SQLFunctionCache *SQLFunctionCachePtr;
 
 /* non-export function prototypes */
 static execution_state *init_execution_state(List *queryTree_list,
-											 bool readonly_func);
+					 bool readonly_func);
 static void init_sql_fcache(FmgrInfo *finfo);
 static void postquel_start(execution_state *es, SQLFunctionCachePtr fcache);
 static TupleTableSlot *postquel_getnext(execution_state *es);
@@ -114,14 +113,14 @@ init_execution_state(List *queryTree_list, bool readonly_func)
 			IsA(queryTree->utilityStmt, TransactionStmt))
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-					 /* translator: %s is a SQL statement name */
+			/* translator: %s is a SQL statement name */
 					 errmsg("%s is not allowed in a SQL function",
 							CreateQueryTag(queryTree))));
 
 		if (readonly_func && !QueryIsReadOnly(queryTree))
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-					 /* translator: %s is a SQL statement name */
+			/* translator: %s is a SQL statement name */
 					 errmsg("%s is not allowed in a non-volatile function",
 							CreateQueryTag(queryTree))));
 
@@ -174,8 +173,8 @@ init_sql_fcache(FmgrInfo *finfo)
 	procedureStruct = (Form_pg_proc) GETSTRUCT(procedureTuple);
 
 	/*
-	 * get the result type from the procedure tuple, and check for
-	 * polymorphic result type; if so, find out the actual result type.
+	 * get the result type from the procedure tuple, and check for polymorphic
+	 * result type; if so, find out the actual result type.
 	 */
 	rettype = procedureStruct->prorettype;
 
@@ -186,7 +185,7 @@ init_sql_fcache(FmgrInfo *finfo)
 			ereport(ERROR,
 					(errcode(ERRCODE_DATATYPE_MISMATCH),
 					 errmsg("could not determine actual result type for function declared to return type %s",
-						  format_type_be(procedureStruct->prorettype))));
+							format_type_be(procedureStruct->prorettype))));
 	}
 
 	fcache->rettype = rettype;
@@ -208,7 +207,7 @@ init_sql_fcache(FmgrInfo *finfo)
 
 		argOidVect = (Oid *) palloc(nargs * sizeof(Oid));
 		memcpy(argOidVect,
-			   procedureStruct->proargtypes,
+			   procedureStruct->proargtypes.values,
 			   nargs * sizeof(Oid));
 		/* Resolve any polymorphic argument types */
 		for (argnum = 0; argnum < nargs; argnum++)
@@ -260,8 +259,8 @@ init_sql_fcache(FmgrInfo *finfo)
 	 * JunkFilter we can use to coerce the returned rowtype to the desired
 	 * form.
 	 */
-	fcache->returnsTuple = check_sql_fn_retval(rettype,
-											   get_typtype(rettype),
+	fcache->returnsTuple = check_sql_fn_retval(foid,
+											   rettype,
 											   queryTree_list,
 											   &fcache->junkFilter);
 
@@ -287,9 +286,9 @@ postquel_start(execution_state *es, SQLFunctionCachePtr fcache)
 	/*
 	 * In a read-only function, use the surrounding query's snapshot;
 	 * otherwise take a new snapshot for each query.  The snapshot should
-	 * include a fresh command ID so that all work to date in this
-	 * transaction is visible.  We copy in both cases so that postquel_end
-	 * can unconditionally do FreeSnapshot.
+	 * include a fresh command ID so that all work to date in this transaction
+	 * is visible.	We copy in both cases so that postquel_end can
+	 * unconditionally do FreeSnapshot.
 	 */
 	if (fcache->readonly_func)
 		snapshot = CopySnapshot(ActiveSnapshot);
@@ -317,7 +316,7 @@ postquel_start(execution_state *es, SQLFunctionCachePtr fcache)
 		 */
 		if (es->qd->operation != CMD_SELECT)
 			AfterTriggerBeginQuery();
-		ExecutorStart(es->qd, false);
+		ExecutorStart(es->qd, 0);
 	}
 
 	es->status = F_EXEC_RUN;
@@ -350,7 +349,9 @@ postquel_getnext(execution_state *es)
 			 * run it to completion.  (If we run to completion then
 			 * ExecutorRun is guaranteed to return NULL.)
 			 */
-			if (LAST_POSTQUEL_COMMAND(es) && es->qd->operation == CMD_SELECT)
+			if (LAST_POSTQUEL_COMMAND(es) &&
+				es->qd->operation == CMD_SELECT &&
+				es->qd->parsetree->into == NULL)
 				count = 1L;
 			else
 				count = 0L;
@@ -388,9 +389,9 @@ postquel_end(execution_state *es)
 		{
 			ActiveSnapshot = es->qd->snapshot;
 
-			ExecutorEnd(es->qd);
 			if (es->qd->operation != CMD_SELECT)
-				AfterTriggerEndQuery();
+				AfterTriggerEndQuery(es->qd->estate);
+			ExecutorEnd(es->qd);
 		}
 		PG_CATCH();
 		{
@@ -419,17 +420,20 @@ postquel_sub_params(SQLFunctionCachePtr fcache,
 	{
 		int			i;
 
-		paramLI = (ParamListInfo) palloc0((nargs + 1) * sizeof(ParamListInfoData));
+		/* sizeof(ParamListInfoData) includes the first array element */
+		paramLI = (ParamListInfo) palloc(sizeof(ParamListInfoData) +
+									   (nargs - 1) *sizeof(ParamExternData));
+		paramLI->numParams = nargs;
 
 		for (i = 0; i < nargs; i++)
 		{
-			paramLI[i].kind = PARAM_NUM;
-			paramLI[i].id = i + 1;
-			paramLI[i].ptype = fcache->argtypes[i];
-			paramLI[i].value = fcinfo->arg[i];
-			paramLI[i].isnull = fcinfo->argnull[i];
+			ParamExternData *prm = &paramLI->params[i];
+
+			prm->value = fcinfo->arg[i];
+			prm->isnull = fcinfo->argnull[i];
+			prm->pflags = 0;
+			prm->ptype = fcache->argtypes[i];
 		}
-		paramLI[nargs].kind = PARAM_INVALID;
 	}
 	else
 		paramLI = NULL;
@@ -447,8 +451,6 @@ postquel_execute(execution_state *es,
 				 MemoryContext resultcontext)
 {
 	TupleTableSlot *slot;
-	HeapTuple	tup;
-	TupleDesc	tupDesc;
 	Datum		value;
 	MemoryContext oldcontext;
 
@@ -460,8 +462,8 @@ postquel_execute(execution_state *es,
 	if (TupIsNull(slot))
 	{
 		/*
-		 * We fall out here for all cases except where we have obtained
-		 * a row from a function's final SELECT.
+		 * We fall out here for all cases except where we have obtained a row
+		 * from a function's final SELECT.
 		 */
 		postquel_end(es);
 		fcinfo->isnull = true;
@@ -469,34 +471,34 @@ postquel_execute(execution_state *es,
 	}
 
 	/*
-	 * If we got a row from a command within the function it has to be
-	 * the final command.  All others shouldn't be returning anything.
+	 * If we got a row from a command within the function it has to be the
+	 * final command.  All others shouldn't be returning anything.
 	 */
 	Assert(LAST_POSTQUEL_COMMAND(es));
 
 	/*
-	 * Set up to return the function value.  For pass-by-reference
-	 * datatypes, be sure to allocate the result in resultcontext,
-	 * not the current memory context (which has query lifespan).
+	 * Set up to return the function value.  For pass-by-reference datatypes,
+	 * be sure to allocate the result in resultcontext, not the current memory
+	 * context (which has query lifespan).
 	 */
 	oldcontext = MemoryContextSwitchTo(resultcontext);
 
 	if (fcache->returnsTuple)
 	{
 		/*
-		 * We are returning the whole tuple, so filter it and apply the
-		 * proper labeling to make it a valid Datum.  There are several
-		 * reasons why we do this:
+		 * We are returning the whole tuple, so filter it and apply the proper
+		 * labeling to make it a valid Datum.  There are several reasons why
+		 * we do this:
 		 *
-		 * 1. To copy the tuple out of the child execution context and
-		 * into the desired result context.
+		 * 1. To copy the tuple out of the child execution context and into
+		 * the desired result context.
 		 *
 		 * 2. To remove any junk attributes present in the raw subselect
-		 * result.  (This is probably not absolutely necessary, but it
-		 * seems like good policy.)
+		 * result. (This is probably not absolutely necessary, but it seems
+		 * like good policy.)
 		 *
-		 * 3. To insert dummy null columns if the declared result type
-		 * has any attisdropped columns.
+		 * 3. To insert dummy null columns if the declared result type has any
+		 * attisdropped columns.
 		 */
 		HeapTuple	newtup;
 		HeapTupleHeader dtup;
@@ -507,19 +509,18 @@ postquel_execute(execution_state *es,
 		newtup = ExecRemoveJunk(fcache->junkFilter, slot);
 
 		/*
-		 * Compress out the HeapTuple header data.  We assume that
-		 * heap_formtuple made the tuple with header and body in one
-		 * palloc'd chunk.  We want to return a pointer to the chunk
-		 * start so that it will work if someone tries to free it.
+		 * Compress out the HeapTuple header data.	We assume that
+		 * heap_form_tuple made the tuple with header and body in one palloc'd
+		 * chunk.  We want to return a pointer to the chunk start so that it
+		 * will work if someone tries to free it.
 		 */
 		t_len = newtup->t_len;
 		dtup = (HeapTupleHeader) newtup;
 		memmove((char *) dtup, (char *) newtup->t_data, t_len);
 
 		/*
-		 * Use the declared return type if it's not RECORD; else take
-		 * the type from the computed result, making sure a typmod has
-		 * been assigned.
+		 * Use the declared return type if it's not RECORD; else take the type
+		 * from the computed result, making sure a typmod has been assigned.
 		 */
 		if (fcache->rettype != RECORDOID)
 		{
@@ -530,7 +531,8 @@ postquel_execute(execution_state *es,
 		else
 		{
 			/* function is declared to return RECORD */
-			tupDesc = fcache->junkFilter->jf_cleanTupType;
+			TupleDesc	tupDesc = fcache->junkFilter->jf_cleanTupType;
+
 			if (tupDesc->tdtypeid == RECORDOID &&
 				tupDesc->tdtypmod < 0)
 				assign_record_type_typmod(tupDesc);
@@ -548,14 +550,10 @@ postquel_execute(execution_state *es,
 	else
 	{
 		/*
-		 * Returning a scalar, which we have to extract from the first
-		 * column of the SELECT result, and then copy into result
-		 * context if needed.
+		 * Returning a scalar, which we have to extract from the first column
+		 * of the SELECT result, and then copy into result context if needed.
 		 */
-		tup = slot->val;
-		tupDesc = slot->ttc_tupleDescriptor;
-
-		value = heap_getattr(tup, 1, tupDesc, &(fcinfo->isnull));
+		value = slot_getattr(slot, 1, &(fcinfo->isnull));
 
 		if (!fcinfo->isnull)
 			value = datumCopy(value, fcache->typbyval, fcache->typlen);
@@ -609,8 +607,8 @@ fmgr_sql(PG_FUNCTION_ARGS)
 	es = fcache->func_state;
 
 	/*
-	 * Convert params to appropriate format if starting a fresh execution.
-	 * (If continuing execution, we can re-use prior params.)
+	 * Convert params to appropriate format if starting a fresh execution. (If
+	 * continuing execution, we can re-use prior params.)
 	 */
 	if (es && es->status == F_EXEC_START)
 		postquel_sub_params(fcache, fcinfo);
@@ -623,8 +621,7 @@ fmgr_sql(PG_FUNCTION_ARGS)
 
 	/*
 	 * Execute each command in the function one after another until we're
-	 * executing the final command and get a result or we run out of
-	 * commands.
+	 * executing the final command and get a result or we run out of commands.
 	 */
 	while (es)
 	{
@@ -683,8 +680,8 @@ fmgr_sql(PG_FUNCTION_ARGS)
 	}
 
 	/*
-	 * If we got a result from a command within the function it has to be
-	 * the final command.  All others shouldn't be returning anything.
+	 * If we got a result from a command within the function it has to be the
+	 * final command.  All others shouldn't be returning anything.
 	 */
 	Assert(LAST_POSTQUEL_COMMAND(es));
 
@@ -703,8 +700,8 @@ fmgr_sql(PG_FUNCTION_ARGS)
 					 errmsg("set-valued function called in context that cannot accept a set")));
 
 		/*
-		 * Ensure we will get shut down cleanly if the exprcontext is not
-		 * run to completion.
+		 * Ensure we will get shut down cleanly if the exprcontext is not run
+		 * to completion.
 		 */
 		if (!fcache->shutdown_reg)
 		{
@@ -746,8 +743,7 @@ sql_exec_error_callback(void *arg)
 	fn_name = NameStr(functup->proname);
 
 	/*
-	 * If there is a syntax error position, convert to internal syntax
-	 * error
+	 * If there is a syntax error position, convert to internal syntax error
 	 */
 	syntaxerrposition = geterrposition();
 	if (syntaxerrposition > 0)
@@ -768,11 +764,11 @@ sql_exec_error_callback(void *arg)
 	}
 
 	/*
-	 * Try to determine where in the function we failed.  If there is a
-	 * query with non-null QueryDesc, finger it.  (We check this rather
-	 * than looking for F_EXEC_RUN state, so that errors during
-	 * ExecutorStart or ExecutorEnd are blamed on the appropriate query;
-	 * see postquel_start and postquel_end.)
+	 * Try to determine where in the function we failed.  If there is a query
+	 * with non-null QueryDesc, finger it.	(We check this rather than looking
+	 * for F_EXEC_RUN state, so that errors during ExecutorStart or
+	 * ExecutorEnd are blamed on the appropriate query; see postquel_start and
+	 * postquel_end.)
 	 */
 	if (fcache)
 	{
@@ -848,9 +844,9 @@ ShutdownSQLFunction(Datum arg)
  * as rettype.  (This means we can't check the type during function definition
  * of a polymorphic function.)
  *
- * The return value is true if the function returns the entire tuple result
- * of its final SELECT, and false otherwise.  Note that because we allow
- * "SELECT rowtype_expression", this may be false even when the declared
+ * This function returns true if the sql function returns the entire tuple
+ * result of its final SELECT, and false otherwise.  Note that because we
+ * allow "SELECT rowtype_expression", this may be false even when the declared
  * function return type is a rowtype.
  *
  * If junkFilter isn't NULL, then *junkFilter is set to a JunkFilter defined
@@ -859,20 +855,15 @@ ShutdownSQLFunction(Datum arg)
  * tuple result), *junkFilter is set to NULL.
  */
 bool
-check_sql_fn_retval(Oid rettype, char fn_typtype, List *queryTreeList,
+check_sql_fn_retval(Oid func_id, Oid rettype, List *queryTreeList,
 					JunkFilter **junkFilter)
 {
 	Query	   *parse;
-	int			cmd;
 	List	   *tlist;
 	ListCell   *tlistitem;
 	int			tlistlen;
-	Oid			typerelid;
+	char		fn_typtype;
 	Oid			restype;
-	Relation	reln;
-	int			relnatts;		/* physical number of columns in rel */
-	int			rellogcols;		/* # of nondeleted columns in rel */
-	int			colindex;		/* physical column index */
 
 	if (junkFilter)
 		*junkFilter = NULL;		/* default result */
@@ -883,103 +874,115 @@ check_sql_fn_retval(Oid rettype, char fn_typtype, List *queryTreeList,
 		if (rettype != VOIDOID)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-					 errmsg("return type mismatch in function declared to return %s",
-							format_type_be(rettype)),
-			 errdetail("Function's final statement must be a SELECT.")));
+			 errmsg("return type mismatch in function declared to return %s",
+					format_type_be(rettype)),
+				 errdetail("Function's final statement must be a SELECT.")));
 		return false;
 	}
 
 	/* find the final query */
 	parse = (Query *) lfirst(list_tail(queryTreeList));
 
-	cmd = parse->commandType;
-	tlist = parse->targetList;
-
 	/*
-	 * The last query must be a SELECT if and only if return type isn't
-	 * VOID.
+	 * If the last query isn't a SELECT, the return type must be VOID.
+	 *
+	 * Note: eventually replace this test with QueryReturnsTuples?  We'd need
+	 * a more general method of determining the output type, though.
 	 */
-	if (rettype == VOIDOID)
+	if (!(parse->commandType == CMD_SELECT && parse->into == NULL))
 	{
-		if (cmd == CMD_SELECT)
+		if (rettype != VOIDOID)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-					 errmsg("return type mismatch in function declared to return %s",
-							format_type_be(rettype)),
-					 errdetail("Function's final statement must not be a SELECT.")));
+			 errmsg("return type mismatch in function declared to return %s",
+					format_type_be(rettype)),
+				 errdetail("Function's final statement must be a SELECT.")));
 		return false;
 	}
 
-	/* by here, the function is declared to return some type */
-	if (cmd != CMD_SELECT)
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-		 errmsg("return type mismatch in function declared to return %s",
-				format_type_be(rettype)),
-			 errdetail("Function's final statement must be a SELECT.")));
+	/*
+	 * OK, it's a SELECT, so it must return something matching the declared
+	 * type.  (We used to insist that the declared type not be VOID in this
+	 * case, but that makes it hard to write a void function that exits
+	 * after calling another void function.  Instead, we insist that the
+	 * SELECT return void ... so void is treated as if it were a scalar type
+	 * below.)
+	 */
 
 	/*
 	 * Count the non-junk entries in the result targetlist.
 	 */
+	tlist = parse->targetList;
 	tlistlen = ExecCleanTargetListLength(tlist);
 
-	typerelid = typeidTypeRelid(rettype);
+	fn_typtype = get_typtype(rettype);
 
-	if (fn_typtype == 'b' || fn_typtype == 'd')
+	if (fn_typtype == 'b' || fn_typtype == 'd' ||
+		rettype == VOIDOID)
 	{
-		/* Shouldn't have a typerelid */
-		Assert(typerelid == InvalidOid);
-
 		/*
-		 * For base-type returns, the target list should have exactly one
-		 * entry, and its type should agree with what the user declared.
-		 * (As of Postgres 7.2, we accept binary-compatible types too.)
+		 * For scalar-type returns, the target list should have exactly one
+		 * entry, and its type should agree with what the user declared. (As
+		 * of Postgres 7.2, we accept binary-compatible types too.)
 		 */
 		if (tlistlen != 1)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-					 errmsg("return type mismatch in function declared to return %s",
-							format_type_be(rettype)),
-			 errdetail("Final SELECT must return exactly one column.")));
+			 errmsg("return type mismatch in function declared to return %s",
+					format_type_be(rettype)),
+				 errdetail("Final SELECT must return exactly one column.")));
 
-		restype = ((TargetEntry *) linitial(tlist))->resdom->restype;
+		restype = exprType((Node *) ((TargetEntry *) linitial(tlist))->expr);
 		if (!IsBinaryCoercible(restype, rettype))
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-					 errmsg("return type mismatch in function declared to return %s",
-							format_type_be(rettype)),
+			 errmsg("return type mismatch in function declared to return %s",
+					format_type_be(rettype)),
 					 errdetail("Actual return type is %s.",
 							   format_type_be(restype))));
 	}
-	else if (fn_typtype == 'c')
+	else if (fn_typtype == 'c' || rettype == RECORDOID)
 	{
-		/* Must have a typerelid */
-		Assert(typerelid != InvalidOid);
+		/* Returns a rowtype */
+		TupleDesc	tupdesc;
+		int			tupnatts;	/* physical number of columns in tuple */
+		int			tuplogcols; /* # of nondeleted columns in tuple */
+		int			colindex;	/* physical column index */
 
 		/*
-		 * If the target list is of length 1, and the type of the varnode
-		 * in the target list matches the declared return type, this is
-		 * okay. This can happen, for example, where the body of the
-		 * function is 'SELECT func2()', where func2 has the same return
-		 * type as the function that's calling it.
+		 * If the target list is of length 1, and the type of the varnode in
+		 * the target list matches the declared return type, this is okay.
+		 * This can happen, for example, where the body of the function is
+		 * 'SELECT func2()', where func2 has the same return type as the
+		 * function that's calling it.
 		 */
 		if (tlistlen == 1)
 		{
-			restype = ((TargetEntry *) linitial(tlist))->resdom->restype;
+			restype = exprType((Node *) ((TargetEntry *) linitial(tlist))->expr);
 			if (IsBinaryCoercible(restype, rettype))
 				return false;	/* NOT returning whole tuple */
 		}
 
+		/* Is the rowtype fixed, or determined only at runtime? */
+		if (get_func_result_type(func_id, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
+		{
+			/*
+			 * Assume we are returning the whole tuple. Crosschecking against
+			 * what the caller expects will happen at runtime.
+			 */
+			if (junkFilter)
+				*junkFilter = ExecInitJunkFilter(tlist, false, NULL);
+			return true;
+		}
+		Assert(tupdesc);
+
 		/*
-		 * Otherwise verify that the targetlist matches the return tuple
-		 * type. This part of the typechecking is a hack. We look up the
-		 * relation that is the declared return type, and scan the
-		 * non-deleted attributes to ensure that they match the datatypes
+		 * Verify that the targetlist matches the return tuple type. We scan
+		 * the non-deleted attributes to ensure that they match the datatypes
 		 * of the non-resjunk columns.
 		 */
-		reln = relation_open(typerelid, AccessShareLock);
-		relnatts = reln->rd_rel->relnatts;
-		rellogcols = 0;			/* we'll count nondeleted cols as we go */
+		tupnatts = tupdesc->natts;
+		tuplogcols = 0;			/* we'll count nondeleted cols as we go */
 		colindex = 0;
 
 		foreach(tlistitem, tlist)
@@ -989,21 +992,21 @@ check_sql_fn_retval(Oid rettype, char fn_typtype, List *queryTreeList,
 			Oid			tletype;
 			Oid			atttype;
 
-			if (tle->resdom->resjunk)
+			if (tle->resjunk)
 				continue;
 
 			do
 			{
 				colindex++;
-				if (colindex > relnatts)
+				if (colindex > tupnatts)
 					ereport(ERROR,
 							(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
 							 errmsg("return type mismatch in function declared to return %s",
 									format_type_be(rettype)),
-					errdetail("Final SELECT returns too many columns.")));
-				attr = reln->rd_att->attrs[colindex - 1];
+					   errdetail("Final SELECT returns too many columns.")));
+				attr = tupdesc->attrs[colindex - 1];
 			} while (attr->attisdropped);
-			rellogcols++;
+			tuplogcols++;
 
 			tletype = exprType((Node *) tle->expr);
 			atttype = attr->atttypid;
@@ -1015,62 +1018,34 @@ check_sql_fn_retval(Oid rettype, char fn_typtype, List *queryTreeList,
 						 errdetail("Final SELECT returns %s instead of %s at column %d.",
 								   format_type_be(tletype),
 								   format_type_be(atttype),
-								   rellogcols)));
+								   tuplogcols)));
 		}
 
 		for (;;)
 		{
 			colindex++;
-			if (colindex > relnatts)
+			if (colindex > tupnatts)
 				break;
-			if (!reln->rd_att->attrs[colindex - 1]->attisdropped)
-				rellogcols++;
+			if (!tupdesc->attrs[colindex - 1]->attisdropped)
+				tuplogcols++;
 		}
 
-		if (tlistlen != rellogcols)
+		if (tlistlen != tuplogcols)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-					 errmsg("return type mismatch in function declared to return %s",
-							format_type_be(rettype)),
+			 errmsg("return type mismatch in function declared to return %s",
+					format_type_be(rettype)),
 					 errdetail("Final SELECT returns too few columns.")));
 
 		/* Set up junk filter if needed */
 		if (junkFilter)
 			*junkFilter = ExecInitJunkFilterConversion(tlist,
-											CreateTupleDescCopy(reln->rd_att),
-											NULL);
-
-		relation_close(reln, AccessShareLock);
+												CreateTupleDescCopy(tupdesc),
+													   NULL);
 
 		/* Report that we are returning entire tuple result */
 		return true;
 	}
-	else if (rettype == RECORDOID)
-	{
-		/*
-		 * If the target list is of length 1, and the type of the varnode
-		 * in the target list matches the declared return type, this is
-		 * okay. This can happen, for example, where the body of the
-		 * function is 'SELECT func2()', where func2 has the same return
-		 * type as the function that's calling it.
-		 */
-		if (tlistlen == 1)
-		{
-			restype = ((TargetEntry *) linitial(tlist))->resdom->restype;
-			if (IsBinaryCoercible(restype, rettype))
-				return false;	/* NOT returning whole tuple */
-		}
-
-		/*
-		 * Otherwise assume we are returning the whole tuple.
-		 * Crosschecking against what the caller expects will happen at
-		 * runtime.
-		 */
-		if (junkFilter)
-			*junkFilter = ExecInitJunkFilter(tlist, false, NULL);
-
-		return true;
-	}
 	else if (rettype == ANYARRAYOID || rettype == ANYELEMENTOID)
 	{
 		/* This should already have been caught ... */
@@ -1082,8 +1057,8 @@ check_sql_fn_retval(Oid rettype, char fn_typtype, List *queryTreeList,
 	else
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-			  errmsg("return type %s is not supported for SQL functions",
-					 format_type_be(rettype))));
+				 errmsg("return type %s is not supported for SQL functions",
+						format_type_be(rettype))));
 
 	return false;
 }
