@@ -3,12 +3,12 @@
  * nodeFunctionscan.c
  *	  Support routines for scanning RangeFunctions (functions in rangetable).
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2006, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/executor/nodeFunctionscan.c,v 1.29 2004/12/31 21:59:45 pgsql Exp $
+ *	  $PostgreSQL: pgsql/src/backend/executor/nodeFunctionscan.c,v 1.40.2.1 2006/12/26 19:26:56 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -22,21 +22,14 @@
  */
 #include "postgres.h"
 
-#include "access/heapam.h"
-#include "catalog/pg_type.h"
-#include "executor/execdebug.h"
-#include "executor/execdefs.h"
-#include "executor/execdesc.h"
 #include "executor/nodeFunctionscan.h"
+#include "funcapi.h"
 #include "parser/parsetree.h"
-#include "parser/parse_expr.h"
-#include "parser/parse_type.h"
-#include "utils/lsyscache.h"
-#include "utils/typcache.h"
+#include "utils/builtins.h"
 
 
 static TupleTableSlot *FunctionNext(FunctionScanState *node);
-static bool tupledesc_match(TupleDesc dst_tupdesc, TupleDesc src_tupdesc);
+static void tupledesc_match(TupleDesc dst_tupdesc, TupleDesc src_tupdesc);
 
 /* ----------------------------------------------------------------
  *						Scan Support
@@ -55,8 +48,6 @@ FunctionNext(FunctionScanState *node)
 	EState	   *estate;
 	ScanDirection direction;
 	Tuplestorestate *tuplestorestate;
-	bool		should_free;
-	HeapTuple	heapTuple;
 
 	/*
 	 * get information from the estate and scan state
@@ -67,9 +58,8 @@ FunctionNext(FunctionScanState *node)
 	tuplestorestate = node->tuplestorestate;
 
 	/*
-	 * If first time through, read all tuples from function and put them
-	 * in a tuplestore. Subsequent calls just fetch tuples from
-	 * tuplestore.
+	 * If first time through, read all tuples from function and put them in a
+	 * tuplestore. Subsequent calls just fetch tuples from tuplestore.
 	 */
 	if (tuplestorestate == NULL)
 	{
@@ -84,23 +74,21 @@ FunctionNext(FunctionScanState *node)
 
 		/*
 		 * If function provided a tupdesc, cross-check it.	We only really
-		 * need to do this for functions returning RECORD, but might as
-		 * well do it always.
+		 * need to do this for functions returning RECORD, but might as well
+		 * do it always.
 		 */
-		if (funcTupdesc && !tupledesc_match(node->tupdesc, funcTupdesc))
-			ereport(ERROR,
-					(errcode(ERRCODE_DATATYPE_MISMATCH),
-					 errmsg("query-specified return row and actual function return row do not match")));
+		if (funcTupdesc)
+			tupledesc_match(node->tupdesc, funcTupdesc);
 	}
 
 	/*
 	 * Get the next tuple from tuplestore. Return NULL if no more tuples.
 	 */
-	heapTuple = tuplestore_getheaptuple(tuplestorestate,
-										ScanDirectionIsForward(direction),
-										&should_free);
 	slot = node->ss.ss_ScanTupleSlot;
-	return ExecStoreTuple(heapTuple, slot, InvalidBuffer, should_free);
+	(void) tuplestore_gettupleslot(tuplestorestate,
+								   ScanDirectionIsForward(direction),
+								   slot);
+	return slot;
 }
 
 /* ----------------------------------------------------------------
@@ -127,7 +115,7 @@ ExecFunctionScan(FunctionScanState *node)
  * ----------------------------------------------------------------
  */
 FunctionScanState *
-ExecInitFunctionScan(FunctionScan *node, EState *estate)
+ExecInitFunctionScan(FunctionScan *node, EState *estate, int eflags)
 {
 	FunctionScanState *scanstate;
 	RangeTblEntry *rte;
@@ -178,18 +166,21 @@ ExecInitFunctionScan(FunctionScan *node, EState *estate)
 	 */
 	rte = rt_fetch(node->scan.scanrelid, estate->es_range_table);
 	Assert(rte->rtekind == RTE_FUNCTION);
-	funcrettype = exprType(rte->funcexpr);
 
 	/*
-	 * Now determine if the function returns a simple or composite type,
-	 * and build an appropriate tupdesc.
+	 * Now determine if the function returns a simple or composite type, and
+	 * build an appropriate tupdesc.
 	 */
-	functypclass = get_type_func_class(funcrettype);
+	functypclass = get_expr_result_type(rte->funcexpr,
+										&funcrettype,
+										&tupdesc);
 
 	if (functypclass == TYPEFUNC_COMPOSITE)
 	{
 		/* Composite data type, e.g. a table's row type */
-		tupdesc = CreateTupleDescCopy(lookup_rowtype_tupdesc(funcrettype, -1));
+		Assert(tupdesc);
+		/* Must copy it out of typcache for safety */
+		tupdesc = CreateTupleDescCopy(tupdesc);
 	}
 	else if (functypclass == TYPEFUNC_SCALAR)
 	{
@@ -206,7 +197,9 @@ ExecInitFunctionScan(FunctionScan *node, EState *estate)
 	}
 	else if (functypclass == TYPEFUNC_RECORD)
 	{
-		tupdesc = BuildDescForRelation(rte->coldeflist);
+		tupdesc = BuildDescFromLists(rte->eref->colnames,
+									 rte->funccoltypes,
+									 rte->funccoltypmods);
 	}
 	else
 	{
@@ -216,15 +209,13 @@ ExecInitFunctionScan(FunctionScan *node, EState *estate)
 
 	/*
 	 * For RECORD results, make sure a typmod has been assigned.  (The
-	 * function should do this for itself, but let's cover things in case
-	 * it doesn't.)
+	 * function should do this for itself, but let's cover things in case it
+	 * doesn't.)
 	 */
-	if (tupdesc->tdtypeid == RECORDOID && tupdesc->tdtypmod < 0)
-		assign_record_type_typmod(tupdesc);
+	BlessTupleDesc(tupdesc);
 
 	scanstate->tupdesc = tupdesc;
-	ExecSetSlotDescriptor(scanstate->ss.ss_ScanTupleSlot,
-						  tupdesc, false);
+	ExecAssignScanType(&scanstate->ss, tupdesc);
 
 	/*
 	 * Other node-specific setup
@@ -239,7 +230,7 @@ ExecInitFunctionScan(FunctionScan *node, EState *estate)
 	 * Initialize result tuple type and projection info.
 	 */
 	ExecAssignResultTypeFromTL(&scanstate->ss.ps);
-	ExecAssignProjectionInfo(&scanstate->ss.ps);
+	ExecAssignScanProjectionInfo(&scanstate->ss);
 
 	return scanstate;
 }
@@ -326,6 +317,7 @@ void
 ExecFunctionReScan(FunctionScanState *node, ExprContext *exprCtxt)
 {
 	ExecClearTuple(node->ss.ps.ps_ResultTupleSlot);
+	node->ss.ps.ps_TupFromTlist = false;
 
 	/*
 	 * If we haven't materialized yet, just return.
@@ -334,10 +326,10 @@ ExecFunctionReScan(FunctionScanState *node, ExprContext *exprCtxt)
 		return;
 
 	/*
-	 * Here we have a choice whether to drop the tuplestore (and recompute
-	 * the function outputs) or just rescan it.  This should depend on
-	 * whether the function expression contains parameters and/or is
-	 * marked volatile.  FIXME soon.
+	 * Here we have a choice whether to drop the tuplestore (and recompute the
+	 * function outputs) or just rescan it.  This should depend on whether the
+	 * function expression contains parameters and/or is marked volatile.
+	 * FIXME soon.
 	 */
 	if (node->ss.ps.chgParam != NULL)
 	{
@@ -349,21 +341,26 @@ ExecFunctionReScan(FunctionScanState *node, ExprContext *exprCtxt)
 }
 
 /*
- * Check that function result tuple type (src_tupdesc) matches or can be
- * considered to match what the query expects (dst_tupdesc).
+ * Check that function result tuple type (src_tupdesc) matches or can
+ * be considered to match what the query expects (dst_tupdesc). If
+ * they don't match, ereport.
  *
  * We really only care about number of attributes and data type.
  * Also, we can ignore type mismatch on columns that are dropped in the
  * destination type, so long as the physical storage matches.  This is
  * helpful in some cases involving out-of-date cached plans.
  */
-static bool
+static void
 tupledesc_match(TupleDesc dst_tupdesc, TupleDesc src_tupdesc)
 {
 	int			i;
 
 	if (dst_tupdesc->natts != src_tupdesc->natts)
-		return false;
+		ereport(ERROR,
+				(errcode(ERRCODE_DATATYPE_MISMATCH),
+				 errmsg("function return row and query-specified return row do not match"),
+				 errdetail("Returned row contains %d attributes, but query expects %d.",
+						   src_tupdesc->natts, dst_tupdesc->natts)));
 
 	for (i = 0; i < dst_tupdesc->natts; i++)
 	{
@@ -373,11 +370,20 @@ tupledesc_match(TupleDesc dst_tupdesc, TupleDesc src_tupdesc)
 		if (dattr->atttypid == sattr->atttypid)
 			continue;			/* no worries */
 		if (!dattr->attisdropped)
-			return false;
+			ereport(ERROR,
+					(errcode(ERRCODE_DATATYPE_MISMATCH),
+					 errmsg("function return row and query-specified return row do not match"),
+					 errdetail("Returned type %s at ordinal position %d, but query expects %s.",
+							   format_type_be(sattr->atttypid),
+							   i + 1,
+							   format_type_be(dattr->atttypid))));
+
 		if (dattr->attlen != sattr->attlen ||
 			dattr->attalign != sattr->attalign)
-			return false;
+			ereport(ERROR,
+					(errcode(ERRCODE_DATATYPE_MISMATCH),
+					 errmsg("function return row and query-specified return row do not match"),
+					 errdetail("Physical storage mismatch on dropped attribute at ordinal position %d.",
+							   i + 1)));
 	}
-
-	return true;
 }
